#! /usr/bin/env python3

"""Installation script for brz.
Run it with
 './setup.py install', or
 './setup.py --help' for more options
"""

import os
import os.path
import sys
import copy
import glob

if sys.version_info < (2, 7):
    sys.stderr.write("[ERROR] Not a supported Python version. Need 2.7+\n")
    sys.exit(1)


try:
    import setuptools
except ImportError:
    sys.stderr.write("[ERROR] Please install setuptools\n")
    sys.exit(1)


# NOTE: The directory containing setup.py, whether run by 'python setup.py' or
# './setup.py' or the equivalent with another path, should always be at the
# start of the path, so this should find the right one...
import breezy

def get_long_description():
    dirname = os.path.dirname(__file__)
    readme = os.path.join(dirname, 'README.rst')
    with open(readme, 'r') as f:
        return f.read()


##
# META INFORMATION FOR SETUP
# see http://docs.python.org/dist/meta-data.html
META_INFO = {
    'name': 'breezy',
    'version': breezy.__version__,
    'maintainer': 'Breezy Developers',
    'maintainer_email': 'team@breezy-vcs.org',
    'url': 'https://www.breezy-vcs.org/',
    'description': 'Friendly distributed version control system',
    'license': 'GNU GPL v2',
    'download_url': 'https://launchpad.net/brz/+download',
    'long_description': get_long_description(),
    'classifiers': [
        'Development Status :: 6 - Mature',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: OS Independent',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Programming Language :: C',
        'Topic :: Software Development :: Version Control',
        ],
    'install_requires': [
        'configobj',
        'six>=1.9.0',
        'patiencediff',
        # Technically, Breezy works without these two dependencies too. But there's
        # no way to enable them by default and let users opt out.
        'fastimport>=0.9.8',
        'dulwich>=0.19.12',
        'dulwich<0.20;python_version<"3.0"',
        ],
    'extras_require': {
        'fastimport': [],
        'git': [],
        'launchpad': ['launchpadlib>=1.6.3'],
        },
    'tests_require': [
        'testtools',
    ],
}

# The list of packages is automatically generated later. Add other things
# that are part of BREEZY here.
BREEZY = {}

PKG_DATA = {
    # install files from selftest suite
    'package_data': {'breezy': ['doc/api/*.txt',
                                'tests/test_patches_data/*',
                                'help_topics/en/*.txt',
                                'tests/ssl_certs/ca.crt',
                                'tests/ssl_certs/server_without_pass.key',
                                'tests/ssl_certs/server_with_pass.key',
                                'tests/ssl_certs/server.crt',
                                ]},
    }
I18N_FILES = []
for filepath in glob.glob("breezy/locale/*/LC_MESSAGES/*.mo"):
    langfile = filepath[len("breezy/locale/"):]
    targetpath = os.path.dirname(os.path.join("share/locale", langfile))
    I18N_FILES.append((targetpath, [filepath]))

def get_breezy_packages():
    """Recurse through the breezy directory, and extract the package names"""

    packages = []
    base_path = os.path.dirname(os.path.abspath(breezy.__file__))
    for root, dirs, files in os.walk(base_path):
        if '__init__.py' in files:
            assert root.startswith(base_path)
            # Get just the path below breezy
            package_path = root[len(base_path):]
            # Remove leading and trailing slashes
            package_path = package_path.strip('\\/')
            if not package_path:
                package_name = 'breezy'
            else:
                package_name = (
                    'breezy.' +
                    package_path.replace('/', '.').replace('\\', '.'))
            packages.append(package_name)
    return sorted(packages)


BREEZY['packages'] = get_breezy_packages()


from setuptools import setup
from distutils.version import LooseVersion
from distutils.command.install_scripts import install_scripts
from distutils.command.install_data import install_data
from distutils.command.build import build

###############################
# Overridden distutils actions
###############################

class my_install_scripts(install_scripts):
    """ Customized install_scripts distutils action.
    Create brz.bat for win32.
    """
    def run(self):
        install_scripts.run(self)   # standard action

        if sys.platform == "win32":
            try:
                scripts_dir = os.path.join(sys.prefix, 'Scripts')
                script_path = self._quoted_path(os.path.join(scripts_dir,
                                                             "brz"))
                python_exe = self._quoted_path(sys.executable)
                batch_str = "@%s %s %%*" % (python_exe, script_path)
                batch_path = os.path.join(self.install_dir, "brz.bat")
                with open(batch_path, "w") as f:
                    f.write(batch_str)
                print(("Created: %s" % batch_path))
            except Exception:
                e = sys.exc_info()[1]
                print(("ERROR: Unable to create %s: %s" % (batch_path, e)))

    def _quoted_path(self, path):
        if ' ' in path:
            return '"' + path + '"'
        else:
            return path
#/class my_install_scripts


class bzr_build(build):
    """Customized build distutils action.
    Generate brz.1.
    """

    sub_commands = build.sub_commands + [
        ('build_mo', lambda _: True),
        ]

    def run(self):
        build.run(self)

        from tools import generate_docs
        generate_docs.main(argv=["brz", "man"])


########################
## Setup
########################

from breezy.bzr_distutils import build_mo

command_classes = {'install_scripts': my_install_scripts,
                   'build': bzr_build,
                   'build_mo': build_mo,
                   }
from distutils import log
from distutils.errors import CCompilerError, DistutilsPlatformError
from distutils.extension import Extension
ext_modules = []
try:
    from Cython.Distutils import build_ext
    from Cython.Compiler.Version import version as cython_version
except ImportError:
    have_cython = False
    # try to build the extension from the prior generated source.
    print("")
    print("The python package 'Cython' is not available."
          " If the .c files are available,")
    print("they will be built,"
          " but modifying the .pyx files will not rebuild them.")
    print("")
    from distutils.command.build_ext import build_ext
else:
    minimum_cython_version = '0.29'
    cython_version_info = LooseVersion(cython_version)
    if cython_version_info < LooseVersion(minimum_cython_version):
        print("Version of Cython is too old. "
              "Current is %s, need at least %s."
              % (cython_version, minimum_cython_version))
        print("If the .c files are available, they will be built,"
              " but modifying the .pyx files will not rebuild them.")
        have_cython = False
    else:
        have_cython = True


class build_ext_if_possible(build_ext):

    user_options = build_ext.user_options + [
        ('allow-python-fallback', None,
         "When an extension cannot be built, allow falling"
         " back to the pure-python implementation.")
        ]

    def initialize_options(self):
        build_ext.initialize_options(self)
        self.allow_python_fallback = False

    def run(self):
        try:
            build_ext.run(self)
        except DistutilsPlatformError:
            e = sys.exc_info()[1]
            if not self.allow_python_fallback:
                log.warn('\n  Cannot build extensions.\n'
                         '  Use "build_ext --allow-python-fallback" to use'
                         ' slower python implementations instead.\n')
                raise
            log.warn(str(e))
            log.warn('\n  Extensions cannot be built.\n'
                     '  Using the slower Python implementations instead.\n')

    def build_extension(self, ext):
        try:
            build_ext.build_extension(self, ext)
        except CCompilerError:
            if not self.allow_python_fallback:
                log.warn('\n  Cannot build extension "%s".\n'
                         '  Use "build_ext --allow-python-fallback" to use'
                         ' slower python implementations instead.\n'
                         % (ext.name,))
                raise
            log.warn('\n  Building of "%s" extension failed.\n'
                     '  Using the slower Python implementation instead.'
                     % (ext.name,))


# Override the build_ext if we have Cython available
command_classes['build_ext'] = build_ext_if_possible
unavailable_files = []


def add_cython_extension(module_name, libraries=None, extra_source=[]):
    """Add a cython module to build.

    This will use Cython to auto-generate the .c file if it is available.
    Otherwise it will fall back on the .c file. If the .c file is not
    available, it will warn, and not add anything.

    You can pass any extra options to Extension through kwargs. One example is
    'libraries = []'.

    :param module_name: The python path to the module. This will be used to
        determine the .pyx and .c files to use.
    """
    path = module_name.replace('.', '/')
    cython_name = path + '.pyx'
    c_name = path + '.c'
    define_macros = []
    if sys.platform == 'win32':
        # cython uses the macro WIN32 to detect the platform, even though it
        # should be using something like _WIN32 or MS_WINDOWS, oh well, we can
        # give it the right value.
        define_macros.append(('WIN32', None))
    if have_cython:
        source = [cython_name]
    else:
        if not os.path.isfile(c_name):
            unavailable_files.append(c_name)
            return
        else:
            source = [c_name]
    source.extend(extra_source)
    include_dirs = ['breezy']
    ext_modules.append(
        Extension(
            module_name, source, define_macros=define_macros,
            libraries=libraries, include_dirs=include_dirs))


add_cython_extension('breezy._simple_set_pyx')
ext_modules.append(Extension('breezy._static_tuple_c',
                             ['breezy/_static_tuple_c.c']))
add_cython_extension('breezy._annotator_pyx')
add_cython_extension('breezy._bencode_pyx')
add_cython_extension('breezy._chunks_to_lines_pyx')
add_cython_extension('breezy.bzr._groupcompress_pyx',
                     extra_source=['breezy/bzr/diff-delta.c'])
add_cython_extension('breezy.bzr._knit_load_data_pyx')
add_cython_extension('breezy._known_graph_pyx')
add_cython_extension('breezy._rio_pyx')
if sys.platform == 'win32':
    add_cython_extension('breezy.bzr._dirstate_helpers_pyx',
                         libraries=['Ws2_32'])
    add_cython_extension('breezy._walkdirs_win32')
else:
    add_cython_extension('breezy.bzr._dirstate_helpers_pyx')
    add_cython_extension('breezy._readdir_pyx')
add_cython_extension('breezy.bzr._chk_map_pyx')
add_cython_extension('breezy.bzr._btree_serializer_pyx')


if unavailable_files:
    print('C extension(s) not found:')
    print(('   %s' % ('\n  '.join(unavailable_files),)))
    print('The python versions will be used instead.')
    print("")


def get_tbzr_py2exe_info(includes, excludes, packages, console_targets,
                         gui_targets, data_files):
    packages.append('tbzrcommands')

    # ModuleFinder can't handle runtime changes to __path__, but
    # win32com uses them.  Hook this in so win32com.shell is found.
    import modulefinder
    import win32com
    import cPickle as pickle
    for p in win32com.__path__[1:]:
        modulefinder.AddPackagePath("win32com", p)
    for extra in ["win32com.shell"]:
        __import__(extra)
        m = sys.modules[extra]
        for p in m.__path__[1:]:
            modulefinder.AddPackagePath(extra, p)

    # TBZR points to the TBZR directory
    tbzr_root = os.environ["TBZR"]

    # Ensure tbreezy itself is on sys.path
    sys.path.append(tbzr_root)

    packages.append("tbreezy")

    # collect up our icons.
    cwd = os.getcwd()
    ico_root = os.path.join(tbzr_root, 'tbreezy', 'resources')
    icos = [] # list of (path_root, relative_ico_path)
    # First always brz's icon and its in the root of the brz tree.
    icos.append(('', 'brz.ico'))
    for root, dirs, files in os.walk(ico_root):
        icos.extend([(ico_root, os.path.join(root, f)[len(ico_root) + 1:])
                     for f in files if f.endswith('.ico')])
    # allocate an icon ID for each file and the full path to the ico
    icon_resources = [(rid, os.path.join(ico_dir, ico_name))
                      for rid, (ico_dir, ico_name) in enumerate(icos)]
    # create a string resource with the mapping.  Might as well save the
    # runtime some effort and write a pickle.
    # Runtime expects unicode objects with forward-slash seps.
    fse = sys.getfilesystemencoding()
    map_items = [(f.replace('\\', '/').decode(fse), rid)
                 for rid, (_, f) in enumerate(icos)]
    ico_map = dict(map_items)
    # Create a new resource type of 'ICON_MAP', and use ID=1
    other_resources = [("ICON_MAP", 1, pickle.dumps(ico_map))]

    excludes.extend("""pywin pywin.dialogs pywin.dialogs.list
                       win32ui crawler.Crawler""".split())

    # tbzrcache executables - a "console" version for debugging and a
    # GUI version that is generally used.
    tbzrcache = dict(
        script=os.path.join(tbzr_root, "scripts", "tbzrcache.py"),
        icon_resources=icon_resources,
        other_resources=other_resources,
    )
    console_targets.append(tbzrcache)

    # Make a windows version which is the same except for the base name.
    tbzrcachew = tbzrcache.copy()
    tbzrcachew["dest_base"] = "tbzrcachew"
    gui_targets.append(tbzrcachew)

    # ditto for the tbzrcommand tool
    tbzrcommand = dict(
        script=os.path.join(tbzr_root, "scripts", "tbzrcommand.py"),
        icon_resources=icon_resources,
        other_resources=other_resources,
    )
    console_targets.append(tbzrcommand)
    tbzrcommandw = tbzrcommand.copy()
    tbzrcommandw["dest_base"] = "tbzrcommandw"
    gui_targets.append(tbzrcommandw)

    # A utility to see python output from both C++ and Python based shell
    # extensions
    tracer = dict(script=os.path.join(tbzr_root, "scripts", "tbzrtrace.py"))
    console_targets.append(tracer)

    # The C++ implemented shell extensions.
    dist_dir = os.path.join(tbzr_root, "shellext", "build")
    data_files.append(('', [os.path.join(dist_dir, 'tbzrshellext_x86.dll')]))
    data_files.append(('', [os.path.join(dist_dir, 'tbzrshellext_x64.dll')]))


def get_qbzr_py2exe_info(includes, excludes, packages, data_files):
    # PyQt4 itself still escapes the plugin detection code for some reason...
    includes.append('PyQt4.QtCore')
    includes.append('PyQt4.QtGui')
    includes.append('PyQt4.QtTest')
    includes.append('sip') # extension module required for Qt.
    packages.append('pygments') # colorizer for qbzr
    packages.append('docutils') # html formatting
    includes.append('win32event')  # for qsubprocess stuff
    # the qt binaries might not be on PATH...
    # They seem to install to a place like C:\Python25\PyQt4\*
    # Which is not the same as C:\Python25\Lib\site-packages\PyQt4
    pyqt_dir = os.path.join(sys.prefix, "PyQt4")
    pyqt_bin_dir = os.path.join(pyqt_dir, "bin")
    if os.path.isdir(pyqt_bin_dir):
        path = os.environ.get("PATH", "")
        if pyqt_bin_dir.lower() not in [p.lower() for p in path.split(os.pathsep)]:
            os.environ["PATH"] = path + os.pathsep + pyqt_bin_dir
    # also add all imageformat plugins to distribution
    # We will look in 2 places, dirname(PyQt4.__file__) and pyqt_dir
    base_dirs_to_check = []
    if os.path.isdir(pyqt_dir):
        base_dirs_to_check.append(pyqt_dir)
    try:
        import PyQt4
    except ImportError:
        pass
    else:
        pyqt4_base_dir = os.path.dirname(PyQt4.__file__)
        if pyqt4_base_dir != pyqt_dir:
            base_dirs_to_check.append(pyqt4_base_dir)
    if not base_dirs_to_check:
        log.warn("Can't find PyQt4 installation -> not including imageformat"
                 " plugins")
    else:
        files = []
        for base_dir in base_dirs_to_check:
            plug_dir = os.path.join(base_dir, 'plugins', 'imageformats')
            if os.path.isdir(plug_dir):
                for fname in os.listdir(plug_dir):
                    # Include plugin dlls, but not debugging dlls
                    fullpath = os.path.join(plug_dir, fname)
                    if fname.endswith('.dll') and not fname.endswith('d4.dll'):
                        files.append(fullpath)
        if files:
            data_files.append(('imageformats', files))
        else:
            log.warn('PyQt4 was found, but we could not find any imageformat'
                     ' plugins. Are you sure your configuration is correct?')


def get_svn_py2exe_info(includes, excludes, packages):
    packages.append('subvertpy')
    packages.append('sqlite3')


def get_git_py2exe_info(includes, excludes, packages):
    packages.append('dulwich')


def get_fastimport_py2exe_info(includes, excludes, packages):
    # This is the python-fastimport package, not to be confused with the
    # brz-fastimport plugin.
    packages.append('fastimport')


if 'bdist_wininst' in sys.argv:
    def find_docs():
        docs = []
        for root, dirs, files in os.walk('doc'):
            r = []
            for f in files:
                if (os.path.splitext(f)[1] in ('.html', '.css', '.png', '.pdf')
                        or f == 'quick-start-summary.svg'):
                    r.append(os.path.join(root, f))
            if r:
                relative = root[4:]
                if relative:
                    target = os.path.join('Doc\\Breezy', relative)
                else:
                    target = 'Doc\\Breezy'
                docs.append((target, r))
        return docs

    # python's distutils-based win32 installer
    ARGS = {'scripts': ['brz', 'tools/win32/brz-win32-bdist-postinstall.py'],
            'ext_modules': ext_modules,
            # help pages
            'data_files': find_docs(),
            # for building cython extensions
            'cmdclass': command_classes,
            }

    ARGS.update(META_INFO)
    ARGS.update(BREEZY)
    PKG_DATA['package_data']['breezy'].append('locale/*/LC_MESSAGES/*.mo')
    ARGS.update(PKG_DATA)

    setup(**ARGS)

elif 'py2exe' in sys.argv:
    # py2exe setup
    import py2exe

    # pick real brz version
    import breezy

    version_number = []
    for i in breezy.version_info[:4]:
        try:
            i = int(i)
        except ValueError:
            i = 0
        version_number.append(str(i))
    version_str = '.'.join(version_number)

    # An override to install_data used only by py2exe builds, which arranges
    # to byte-compile any .py files in data_files (eg, our plugins)
    # Necessary as we can't rely on the user having the relevant permissions
    # to the "Program Files" directory to generate them on the fly.
    class install_data_with_bytecompile(install_data):
        def run(self):
            from distutils.util import byte_compile

            install_data.run(self)

            py2exe = self.distribution.get_command_obj('py2exe', False)
            # GZ 2010-04-19: Setup has py2exe.optimize as 2, but give plugins
            #                time before living with docstring stripping
            optimize = 1
            compile_names = [f for f in self.outfiles if f.endswith('.py')]
            # Round mtime to nearest even second so that installing on a FAT
            # filesystem bytecode internal and script timestamps will match
            for f in compile_names:
                mtime = os.stat(f).st_mtime
                remainder = mtime % 2
                if remainder:
                    mtime -= remainder
                    os.utime(f, (mtime, mtime))
            byte_compile(compile_names,
                         optimize=optimize,
                         force=self.force, prefix=self.install_dir,
                         dry_run=self.dry_run)
            self.outfiles.extend([f + 'o' for f in compile_names])
    # end of class install_data_with_bytecompile

    target = py2exe.build_exe.Target(
        script="brz",
        dest_base="brz",
        icon_resources=[(0, 'brz.ico')],
        name=META_INFO['name'],
        version=version_str,
        description=META_INFO['description'],
        author=META_INFO['author'],
        copyright="(c) Canonical Ltd, 2005-2010",
        company_name="Canonical Ltd.",
        comments=META_INFO['description'],
    )
    gui_target = copy.copy(target)
    gui_target.dest_base = "bzrw"

    packages = BREEZY['packages']
    packages.remove('breezy')
    packages = [i for i in packages if not i.startswith('breezy.plugins')]
    includes = []
    for i in glob.glob('breezy\\*.py'):
        module = i[:-3].replace('\\', '.')
        if module.endswith('__init__'):
            module = module[:-len('__init__')]
        includes.append(module)

    additional_packages = set()
    if sys.version.startswith('2.7'):
        additional_packages.add('xml.etree')
    else:
        import warnings
        warnings.warn('Unknown Python version.\n'
                      'Please check setup.py script for compatibility.')

    # Although we currently can't enforce it, we consider it an error for
    # py2exe to report any files are "missing".  Such modules we know aren't
    # used should be listed here.
    excludes = """Tkinter psyco ElementPath r_hmac
                  ImaginaryModule cElementTree elementtree.ElementTree
                  Crypto.PublicKey._fastmath
                  tools
                  resource validate""".split()
    dll_excludes = []

    # email package from std python library use lazy import,
    # so we need to explicitly add all package
    additional_packages.add('email')
    # And it uses funky mappings to conver to 'Oldname' to 'newname'.  As
    # a result, packages like 'email.Parser' show as missing.  Tell py2exe
    # to exclude them.
    import email
    for oldname in getattr(email, '_LOWERNAMES', []):
        excludes.append("email." + oldname)
    for oldname in getattr(email, '_MIMENAMES', []):
        excludes.append("email.MIME" + oldname)

    # text files for help topis
    text_topics = glob.glob('breezy/help_topics/en/*.txt')
    topics_files = [('lib/help_topics/en', text_topics)]

    # built-in plugins
    plugins_files = []
    # XXX - should we consider having the concept of an 'official' build,
    # which hard-codes the list of plugins, gets more upset if modules are
    # missing, etc?
    plugins = None # will be a set after plugin sniffing...
    for root, dirs, files in os.walk('breezy/plugins'):
        if root == 'breezy/plugins':
            plugins = set(dirs)
            # We ship plugins as normal files on the file-system - however,
            # the build process can cause *some* of these plugin files to end
            # up in library.zip. Thus, we saw (eg) "plugins/svn/test" in
            # library.zip, and then saw import errors related to that as the
            # rest of the svn plugin wasn't. So we tell py2exe to leave the
            # plugins out of the .zip file
            excludes.extend(["breezy.plugins." + d for d in dirs])
        x = []
        for i in files:
            # Throw away files we don't want packaged. Note that plugins may
            # have data files with all sorts of extensions so we need to
            # be conservative here about what we ditch.
            ext = os.path.splitext(i)[1]
            if ext.endswith('~') or ext in [".pyc", ".swp"]:
                continue
            if i == '__init__.py' and root == 'breezy/plugins':
                continue
            x.append(os.path.join(root, i))
        if x:
            target_dir = root[len('breezy/'):]  # install to 'plugins/...'
            plugins_files.append((target_dir, x))
    # find modules for built-in plugins
    import tools.package_mf
    mf = tools.package_mf.CustomModuleFinder()
    mf.run_package('breezy/plugins')
    packs, mods = mf.get_result()
    additional_packages.update(packs)
    includes.extend(mods)

    console_targets = [target,
                       'tools/win32/bzr_postinstall.py',
                       ]
    gui_targets = [gui_target]
    data_files = topics_files + plugins_files + I18N_FILES

    if 'qbzr' in plugins:
        get_qbzr_py2exe_info(includes, excludes, packages, data_files)

    if 'svn' in plugins:
        get_svn_py2exe_info(includes, excludes, packages)

    if 'git' in plugins:
        get_git_py2exe_info(includes, excludes, packages)

    if 'fastimport' in plugins:
        get_fastimport_py2exe_info(includes, excludes, packages)

    if "TBZR" in os.environ:
        # TORTOISE_OVERLAYS_MSI_WIN32 must be set to the location of the
        # TortoiseOverlays MSI installer file. It is in the TSVN svn repo and
        # can be downloaded from (username=guest, blank password):
        # http://tortoisesvn.tigris.org/svn/tortoisesvn/TortoiseOverlays
        # look for: version-1.0.4/bin/TortoiseOverlays-1.0.4.11886-win32.msi
        # Ditto for TORTOISE_OVERLAYS_MSI_X64, pointing at *-x64.msi.
        for needed in ('TORTOISE_OVERLAYS_MSI_WIN32',
                       'TORTOISE_OVERLAYS_MSI_X64'):
            url = ('http://guest:@tortoisesvn.tigris.org/svn/tortoisesvn'
                   '/TortoiseOverlays')
            if not os.path.isfile(os.environ.get(needed, '<nofile>')):
                raise RuntimeError(
                    "\nPlease set %s to the location of the relevant"
                    "\nTortoiseOverlays .msi installer file."
                    " The installers can be found at"
                    "\n  %s"
                    "\ncheck in the version-X.Y.Z/bin/ subdir" % (needed, url))
        get_tbzr_py2exe_info(includes, excludes, packages, console_targets,
                             gui_targets, data_files)
    else:
        # print this warning to stderr as output is redirected, so it is seen
        # at build time.  Also to stdout so it appears in the log
        for f in (sys.stderr, sys.stdout):
            f.write("Skipping TBZR binaries - "
                    "please set TBZR to a directory to enable\n")

    # MSWSOCK.dll is a system-specific library, which py2exe accidentally pulls
    # in on Vista.
    dll_excludes.extend(["MSWSOCK.dll",
                         "MSVCP60.dll",
                         "MSVCP90.dll",
                         "powrprof.dll",
                         "SHFOLDER.dll"])
    options_list = {"py2exe": {"packages": packages + list(additional_packages),
                               "includes": includes,
                               "excludes": excludes,
                               "dll_excludes": dll_excludes,
                               "dist_dir": "win32_bzr.exe",
                               "optimize": 2,
                               "custom_boot_script":
                                   "tools/win32/py2exe_boot_common.py",
                               },
                    }

    # We want the libaray.zip to have optimize = 2, but the exe to have
    # optimize = 1, so that .py files that get compilied at run time
    # (e.g. user installed plugins) dont have their doc strings removed.
    class py2exe_no_oo_exe(py2exe.build_exe.py2exe):
        def build_executable(self, *args, **kwargs):
            self.optimize = 1
            py2exe.build_exe.py2exe.build_executable(self, *args, **kwargs)
            self.optimize = 2

    if __name__ == '__main__':
        command_classes['install_data'] = install_data_with_bytecompile
        command_classes['py2exe'] = py2exe_no_oo_exe
        setup(options=options_list,
              console=console_targets,
              windows=gui_targets,
              zipfile='lib/library.zip',
              data_files=data_files,
              cmdclass=command_classes,
              )

else:
    # ad-hoc for easy_install
    DATA_FILES = []
    if 'bdist_egg' not in sys.argv:
        # generate and install brz.1 only with plain install, not the
        # easy_install one
        DATA_FILES = [('man/man1', ['brz.1', 'breezy/git/git-remote-bzr.1'])]

    DATA_FILES = DATA_FILES + I18N_FILES
    # std setup
    ARGS = {'scripts': ['brz',
                        # TODO(jelmer): Only install the git scripts if
                        # Dulwich was found.
                        'breezy/git/git-remote-bzr',
                        'breezy/git/bzr-receive-pack',
                        'breezy/git/bzr-upload-pack'],
            'data_files': DATA_FILES,
            'cmdclass': command_classes,
            'ext_modules': ext_modules,
            }

    ARGS.update(META_INFO)
    ARGS.update(BREEZY)
    ARGS.update(PKG_DATA)

    if __name__ == '__main__':
        setup(**ARGS)



#!/usr/bin/python3

# Copyright 2009 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import os
import sys
import tarfile
from optparse import OptionParser
from shutil import copy2, copytree, rmtree

sys.path.insert(0, os.path.dirname(os.path.dirname(__file__)))


def package_docs(section, src_build, dest_html, dest_downloads):
    """Package docs from a Sphinx _build directory into target directories.

    :param section: section in the website being built
    :param src_build: the _build directory
    :param dest_html: the directory where html should go
    :param downloads: the directory where downloads should go
    """
    # Copy across the HTML. Explicitly delete the html destination
    # directory first though because copytree insists on it not existing.
    src_html = os.path.join(src_build, 'html')
    if os.path.exists(dest_html):
        rmtree(dest_html)
    copytree(src_html, dest_html)

    # Package the html as a downloadable archive
    archive_root = "brz-%s-html" % (section,)
    archive_basename = "%s.tar.bz2" % (archive_root,)
    archive_name = os.path.join(dest_downloads, archive_basename)
    build_archive(src_html, archive_name, archive_root, 'bz2')

    # Copy across the PDF docs, if any, including the quick ref card
    pdf_files = []
    quick_ref = os.path.join(
        src_html, '_static/%s/brz-%s-quick-reference.pdf' % (section, section))
    if os.path.exists(quick_ref):
        pdf_files.append(quick_ref)
    src_pdf = os.path.join(src_build, 'latex')
    if os.path.exists(src_pdf):
        for name in os.listdir(src_pdf):
            if name.endswith('.pdf'):
                pdf_files.append(os.path.join(src_pdf, name))
    if pdf_files:
        dest_pdf = os.path.join(dest_downloads, 'pdf-%s' % (section,))
        if not os.path.exists(dest_pdf):
            os.mkdir(dest_pdf)
        for pdf in pdf_files:
            copy2(pdf, dest_pdf)

    # TODO: copy across the CHM files, if any


def build_archive(src_dir, archive_name, archive_root, format):
    print("creating %s ..." % (archive_name,))
    tar = tarfile.open(archive_name, "w:%s" % (format,))
    for relpath in os.listdir(src_dir):
        src_path = os.path.join(src_dir, relpath)
        archive_path = os.path.join(archive_root, relpath)
        tar.add(src_path, arcname=archive_path)
    tar.close()


def main(argv):
    # Check usage. The first argument is the parent directory of
    # the Sphinx _build directory. It will typically be 'doc/xx'.
    # The second argument is the website build directory.
    parser = OptionParser(usage="%prog SOURCE-DIR WEBSITE-BUILD-DIR")
    (options, args) = parser.parse_args(argv)
    if len(args) != 2:
        parser.print_help()
        sys.exit(1)

    # Get the section - locale code or 'developers'
    src_dir = args[0]
    section = os.path.basename(src_dir)
    src_build = os.path.join(src_dir, '_build')

    # Create the destination directories if they doesn't exist.
    dest_dir = args[1]
    dest_html = os.path.join(dest_dir, section)
    dest_downloads = os.path.join(dest_dir, 'downloads')
    for d in [dest_dir, dest_downloads]:
        if not os.path.exists(d):
            print("creating directory %s ..." % (d,))
            os.mkdir(d)

    # Package and copy the files across
    package_docs(section, src_build, dest_html, dest_downloads)


if __name__ == '__main__':
    main(sys.argv[1:])



# Copyright (C) 2006-2010 Canonical Ltd
# Copyright (C) 2018 Breezy Developers
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""UI location string handling."""

from __future__ import absolute_import

import re

from . import (
    urlutils,
    )
from .hooks import Hooks
from .sixish import (
    PY3,
    string_types,
    )


class LocationHooks(Hooks):
    """Dictionary mapping hook name to a list of callables for location hooks.
    """

    def __init__(self):
        Hooks.__init__(self, "breezy.location", "hooks")
        self.add_hook(
            'rewrite_url',
            "Possibly rewrite a URL. Called with a URL to rewrite and the "
            "purpose of the URL.", (3, 0))


hooks = LocationHooks()


def rcp_location_to_url(location, scheme='ssh'):
    """Convert a rcp-style location to a URL.

    :param location: Location to convert, e.g. "foo:bar"
    :param schenme: URL scheme to return, defaults to "ssh"
    :return: A URL, e.g. "ssh://foo/bar"
    :raises ValueError: if this is not a RCP-style URL
    """
    m = re.match('^(?P<user>[^@:/]+@)?(?P<host>[^/:]+):(?P<path>.*)$', location)
    if not m:
        raise ValueError("Not a RCP URL")
    if m.group('path').startswith('//'):
        raise ValueError("Not a RCP URL: already looks like a URL")
    quoted_user = urlutils.quote(m.group('user')[:-1]) if m.group('user') else None
    url = urlutils.URL(
        scheme=scheme, quoted_user=quoted_user,
        port=None, quoted_password=None,
        quoted_host=urlutils.quote(m.group('host')),
        quoted_path=urlutils.quote(m.group('path')))
    return str(url)


def pserver_to_url(location):
    """Convert a CVS pserver location string to a URL.

    :param location: pserver URL
    :return: A cvs+pserver URL
    """
    parts = location.split(':')
    if parts[0] or parts[1] != 'pserver':
        raise ValueError('not a valid pserver location string')
    try:
        (username, hostname) = parts[2].split('@', 1)
    except IndexError:
        hostname = parts[2]
        username = None
    return str(urlutils.URL(
        scheme='cvs+pserver',
        quoted_user=urlutils.quote(username) if username else None,
        quoted_host=urlutils.quote(hostname),
        quoted_password=None,
        port=None,
        quoted_path=urlutils.quote(parts[3])))


def location_to_url(location, purpose=None):
    """Determine a fully qualified URL from a location string.

    This will try to interpret location as both a URL and a directory path. It
    will also lookup the location in directories.

    :param location: Unicode or byte string object with a location
    :param purpose: Intended method of access (None, 'read' or 'write')
    :raise InvalidURL: If the location is already a URL, but not valid.
    :return: Byte string with resulting URL
    """
    if not isinstance(location, string_types):
        raise AssertionError("location not a byte or unicode string")

    if location.startswith(':pserver:'):
        return pserver_to_url(location)

    from .directory_service import directories
    location = directories.dereference(location, purpose)

    # Catch any URLs which are passing Unicode rather than ASCII
    try:
        location = location.encode('ascii')
    except UnicodeError:
        if urlutils.is_url(location):
            raise urlutils.InvalidURL(
                path=location, extra='URLs must be properly escaped')
        location = urlutils.local_path_to_url(location)
    else:
        if PY3:
            location = location.decode('ascii')

    if location.startswith("file:") and not location.startswith("file://"):
        return urlutils.join(urlutils.local_path_to_url("."), location[5:])

    try:
        url = rcp_location_to_url(location, scheme="ssh")
    except ValueError:
        pass
    else:
        return url

    if not urlutils.is_url(location):
        return urlutils.local_path_to_url(location)

    for hook in hooks['rewrite_url']:
        location = hook(location, purpose=purpose)

    return location



# Copyright (C) 2009, 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Pyrex implementation of _read_stanza_*."""

from __future__ import absolute_import


cdef extern from "python-compat.h":
    pass

from cpython.bytes cimport (
    PyBytes_CheckExact,
    PyBytes_FromStringAndSize,
    PyBytes_AS_STRING,
    PyBytes_GET_SIZE,
    )
from cpython.unicode cimport (
    PyUnicode_CheckExact,
    PyUnicode_DecodeUTF8,
    # Deprecated after PEP 393 changes
    PyUnicode_AS_UNICODE,
    PyUnicode_FromUnicode,
    PyUnicode_GET_SIZE,
    )
from cpython.list cimport (
    PyList_Append,
    )
from cpython.mem cimport (
    PyMem_Free,
    PyMem_Malloc,
    PyMem_Realloc,
    )
from cpython.version cimport (
    PY_MAJOR_VERSION,
    )

cdef extern from "Python.h":
    ctypedef int Py_UNICODE
    object PyUnicode_EncodeASCII(Py_UNICODE *, int, char *)
    int Py_UNICODE_ISLINEBREAK(Py_UNICODE)

    # GZ 2017-09-11: Not sure why cython unicode module lacks this?
    object PyUnicode_FromStringAndSize(const char *u, Py_ssize_t size)

    # Python 3.3 or later unicode handling
    char* PyUnicode_AsUTF8AndSize(object unicode, Py_ssize_t *size)

from libc.string cimport (
    memcpy,
    )

from .rio import Stanza


cdef int _valid_tag_char(char c): # cannot_raise
    return (c == c'_' or c == c'-' or
            (c >= c'a' and c <= c'z') or
            (c >= c'A' and c <= c'Z') or
            (c >= c'0' and c <= c'9'))


def _valid_tag(tag):
    cdef char *c_tag
    cdef Py_ssize_t c_len
    cdef int i
    # GZ 2017-09-11: Encapsulate native string as ascii tag somewhere neater
    if PY_MAJOR_VERSION >= 3:
        if not PyUnicode_CheckExact(tag):
            raise TypeError(tag)
        c_tag = PyUnicode_AsUTF8AndSize(tag, &c_len)
    else:
        if not PyBytes_CheckExact(tag):
            raise TypeError(tag)
        c_tag = PyBytes_AS_STRING(tag)
        c_len = PyBytes_GET_SIZE(tag)
    if c_len < 1:
        return False
    for i from 0 <= i < c_len:
        if not _valid_tag_char(c_tag[i]):
            return False
    return True


cdef object _split_first_line_utf8(char *line, int len,
                                   char *value, Py_ssize_t *value_len):
    cdef int i
    for i from 0 <= i < len:
        if line[i] == c':':
            if line[i+1] != c' ':
                raise ValueError("invalid tag in line %r" % line)
            memcpy(value, line+i+2, len-i-2)
            value_len[0] = len-i-2
            if PY_MAJOR_VERSION >= 3:
                return PyUnicode_FromStringAndSize(line, i)
            return PyBytes_FromStringAndSize(line, i)
    raise ValueError('tag/value separator not found in line %r' % line)


cdef object _split_first_line_unicode(Py_UNICODE *line, int len,
                                      Py_UNICODE *value, Py_ssize_t *value_len):
    cdef int i
    for i from 0 <= i < len:
        if line[i] == c':':
            if line[i+1] != c' ':
                raise ValueError("invalid tag in line %r" %
                                 PyUnicode_FromUnicode(line, len))
            memcpy(value, &line[i+2], (len-i-2) * sizeof(Py_UNICODE))
            value_len[0] = len-i-2
            if PY_MAJOR_VERSION >= 3:
                return PyUnicode_FromUnicode(line, i)
            return PyUnicode_EncodeASCII(line, i, "strict")
    raise ValueError("tag/value separator not found in line %r" %
                     PyUnicode_FromUnicode(line, len))


def _read_stanza_utf8(line_iter):
    cdef char *c_line
    cdef Py_ssize_t c_len
    cdef char *accum_value
    cdef char *new_accum_value
    cdef Py_ssize_t accum_len, accum_size
    pairs = []
    tag = None
    accum_len = 0
    accum_size = 4096
    accum_value = <char *>PyMem_Malloc(accum_size)
    if accum_value == NULL:
        raise MemoryError
    try:
        for line in line_iter:
            if line is None:
                break # end of file
            if not PyBytes_CheckExact(line):
                raise TypeError("%r is not a plain string" % line)
            c_line = PyBytes_AS_STRING(line)
            c_len = PyBytes_GET_SIZE(line)
            if c_len < 1:
                break       # end of file
            if c_len == 1 and c_line[0] == c"\n":
                break       # end of stanza
            if accum_len + c_len > accum_size:
                accum_size = (accum_len + c_len)
                new_accum_value = <char *>PyMem_Realloc(accum_value, accum_size)
                if new_accum_value == NULL:
                    raise MemoryError
                else:
                    accum_value = new_accum_value
            if c_line[0] == c'\t': # continues previous value
                if tag is None:
                    raise ValueError('invalid continuation line %r' % line)
                memcpy(accum_value+accum_len, c_line+1, c_len-1)
                accum_len = accum_len + c_len-1
            else: # new tag:value line
                if tag is not None:
                    PyList_Append(pairs,
                        (tag, PyUnicode_DecodeUTF8(accum_value, accum_len-1,
                                                   "strict")))
                tag = _split_first_line_utf8(c_line, c_len, accum_value,
                                             &accum_len)
                if not _valid_tag(tag):
                    raise ValueError("invalid rio tag %r" % (tag,))
        if tag is not None: # add last tag-value
            PyList_Append(pairs,
                (tag, PyUnicode_DecodeUTF8(accum_value, accum_len-1, "strict")))
            return Stanza.from_pairs(pairs)
        else:     # didn't see any content
            return None
    finally:
        PyMem_Free(accum_value)


def _read_stanza_unicode(unicode_iter):
    cdef Py_UNICODE *c_line
    cdef int c_len
    cdef Py_UNICODE *accum_value
    cdef Py_UNICODE *new_accum_value
    cdef Py_ssize_t accum_len, accum_size
    pairs = []
    tag = None
    accum_len = 0
    accum_size = 4096
    accum_value = <Py_UNICODE *>PyMem_Malloc(accum_size*sizeof(Py_UNICODE))
    if accum_value == NULL:
        raise MemoryError
    try:
        for line in unicode_iter:
            if line is None:
                break       # end of file
            if not PyUnicode_CheckExact(line):
                raise TypeError("%r is not a unicode string" % line)
            c_line = PyUnicode_AS_UNICODE(line)
            c_len = PyUnicode_GET_SIZE(line)
            if c_len < 1:
                break        # end of file
            if Py_UNICODE_ISLINEBREAK(c_line[0]):
                break       # end of stanza
            if accum_len + c_len > accum_size:
                accum_size = accum_len + c_len
                new_accum_value = <Py_UNICODE *>PyMem_Realloc(accum_value,
                    accum_size*sizeof(Py_UNICODE))
                if new_accum_value == NULL:
                    raise MemoryError
                else:
                    accum_value = new_accum_value
            if c_line[0] == c'\t': # continues previous value,
                if tag is None:
                    raise ValueError('invalid continuation line %r' % line)
                memcpy(&accum_value[accum_len], &c_line[1],
                    (c_len-1)*sizeof(Py_UNICODE))
                accum_len = accum_len + (c_len-1)
            else: # new tag:value line
                if tag is not None:
                    PyList_Append(pairs,
                        (tag, PyUnicode_FromUnicode(accum_value, accum_len-1)))
                tag = _split_first_line_unicode(c_line, c_len, accum_value,
                                                &accum_len)
                if not _valid_tag(tag):
                    raise ValueError("invalid rio tag %r" % (tag,))
        if tag is not None: # add last tag-value
            PyList_Append(pairs,
                    (tag, PyUnicode_FromUnicode(accum_value, accum_len-1)))
            return Stanza.from_pairs(pairs)
        else:     # didn't see any content
            return None
    finally:
        PyMem_Free(accum_value)



# Copyright (C) 2009, 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for smart server request infrastructure (breezy.bzr.smart.request)."""

import threading

from breezy import (
    errors,
    transport,
    )
from breezy.bzr.bzrdir import BzrDir
from breezy.bzr.smart import request
from breezy.tests import TestCase, TestCaseWithMemoryTransport


class NoBodyRequest(request.SmartServerRequest):
    """A request that does not implement do_body."""

    def do(self):
        return request.SuccessfulSmartServerResponse(('ok',))


class DoErrorRequest(request.SmartServerRequest):
    """A request that raises an error from self.do()."""

    def do(self):
        raise errors.NoSuchFile('xyzzy')


class DoUnexpectedErrorRequest(request.SmartServerRequest):
    """A request that encounters a generic error in self.do()"""

    def do(self):
        dict()[1]


class ChunkErrorRequest(request.SmartServerRequest):
    """A request that raises an error from self.do_chunk()."""

    def do(self):
        """No-op."""
        pass

    def do_chunk(self, bytes):
        raise errors.NoSuchFile('xyzzy')


class EndErrorRequest(request.SmartServerRequest):
    """A request that raises an error from self.do_end()."""

    def do(self):
        """No-op."""
        pass

    def do_chunk(self, bytes):
        """No-op."""
        pass

    def do_end(self):
        raise errors.NoSuchFile('xyzzy')


class CheckJailRequest(request.SmartServerRequest):

    def __init__(self, *args):
        request.SmartServerRequest.__init__(self, *args)
        self.jail_transports_log = []

    def do(self):
        self.jail_transports_log.append(request.jail_info.transports)

    def do_chunk(self, bytes):
        self.jail_transports_log.append(request.jail_info.transports)

    def do_end(self):
        self.jail_transports_log.append(request.jail_info.transports)


class TestErrors(TestCase):

    def test_disabled_method(self):
        error = request.DisabledMethod("class name")
        self.assertEqualDiff(
            "The smart server method 'class name' is disabled.", str(error))


class TestSmartRequest(TestCase):

    def test_request_class_without_do_body(self):
        """If a request has no body data, and the request's implementation does
        not override do_body, then no exception is raised.
        """
        # Create a SmartServerRequestHandler with a SmartServerRequest subclass
        # that does not implement do_body.
        handler = request.SmartServerRequestHandler(
            None, {b'foo': NoBodyRequest}, '/')
        # Emulate a request with no body (i.e. just args).
        handler.args_received((b'foo',))
        handler.end_received()
        # Request done, no exception was raised.

    def test_only_request_code_is_jailed(self):
        transport = 'dummy transport'
        handler = request.SmartServerRequestHandler(
            transport, {b'foo': CheckJailRequest}, '/')
        handler.args_received((b'foo',))
        self.assertEqual(None, request.jail_info.transports)
        handler.accept_body(b'bytes')
        self.assertEqual(None, request.jail_info.transports)
        handler.end_received()
        self.assertEqual(None, request.jail_info.transports)
        self.assertEqual(
            [[transport]] * 3, handler._command.jail_transports_log)

    def test_all_registered_requests_are_safety_qualified(self):
        unclassified_requests = []
        allowed_info = ('read', 'idem', 'mutate', 'semivfs', 'semi', 'stream')
        for key in request.request_handlers.keys():
            info = request.request_handlers.get_info(key)
            if info is None or info not in allowed_info:
                unclassified_requests.append(key)
        if unclassified_requests:
            self.fail('These requests were not categorized as safe/unsafe'
                      ' to retry: %s' % (unclassified_requests,))


class TestSmartRequestHandlerErrorTranslation(TestCase):
    """Tests that SmartServerRequestHandler will translate exceptions raised by
    a SmartServerRequest into FailedSmartServerResponses.
    """

    def assertNoResponse(self, handler):
        self.assertEqual(None, handler.response)

    def assertResponseIsTranslatedError(self, handler):
        expected_translation = (b'NoSuchFile', b'xyzzy')
        self.assertEqual(
            request.FailedSmartServerResponse(expected_translation),
            handler.response)

    def test_error_translation_from_args_received(self):
        handler = request.SmartServerRequestHandler(
            None, {b'foo': DoErrorRequest}, '/')
        handler.args_received((b'foo',))
        self.assertResponseIsTranslatedError(handler)

    def test_error_translation_from_chunk_received(self):
        handler = request.SmartServerRequestHandler(
            None, {b'foo': ChunkErrorRequest}, '/')
        handler.args_received((b'foo',))
        self.assertNoResponse(handler)
        handler.accept_body(b'bytes')
        self.assertResponseIsTranslatedError(handler)

    def test_error_translation_from_end_received(self):
        handler = request.SmartServerRequestHandler(
            None, {b'foo': EndErrorRequest}, '/')
        handler.args_received((b'foo',))
        self.assertNoResponse(handler)
        handler.end_received()
        self.assertResponseIsTranslatedError(handler)

    def test_unexpected_error_translation(self):
        handler = request.SmartServerRequestHandler(
            None, {b'foo': DoUnexpectedErrorRequest}, '/')
        handler.args_received((b'foo',))
        self.assertEqual(
            request.FailedSmartServerResponse((b'error', b'KeyError', b"1")),
            handler.response)


class TestRequestHanderErrorTranslation(TestCase):
    """Tests for breezy.bzr.smart.request._translate_error."""

    def assertTranslationEqual(self, expected_tuple, error):
        self.assertEqual(expected_tuple, request._translate_error(error))

    def test_NoSuchFile(self):
        self.assertTranslationEqual(
            (b'NoSuchFile', b'path'), errors.NoSuchFile('path'))

    def test_LockContention(self):
        # For now, LockContentions are always transmitted with no details.
        # Eventually they should include a relpath or url or something else to
        # identify which lock is busy.
        self.assertTranslationEqual(
            (b'LockContention',), errors.LockContention('lock', 'msg'))

    def test_TokenMismatch(self):
        self.assertTranslationEqual(
            (b'TokenMismatch', b'some-token', b'actual-token'),
            errors.TokenMismatch(b'some-token', b'actual-token'))

    def test_MemoryError(self):
        self.assertTranslationEqual((b"MemoryError",), MemoryError())

    def test_GhostRevisionsHaveNoRevno(self):
        self.assertTranslationEqual(
            (b"GhostRevisionsHaveNoRevno", b'revid1', b'revid2'),
            errors.GhostRevisionsHaveNoRevno(b'revid1', b'revid2'))

    def test_generic_Exception(self):
        self.assertTranslationEqual((b'error', b'Exception', b""),
                                    Exception())

    def test_generic_BzrError(self):
        self.assertTranslationEqual((b'error', b'BzrError', b"some text"),
                                    errors.BzrError(msg="some text"))

    def test_generic_zlib_error(self):
        from zlib import error
        msg = "Error -3 while decompressing data: incorrect data check"
        self.assertTranslationEqual((b'error', b'zlib.error', msg.encode('utf-8')),
                                    error(msg))


class TestRequestJail(TestCaseWithMemoryTransport):

    def test_jail(self):
        transport = self.get_transport('blah')
        req = request.SmartServerRequest(transport)
        self.assertEqual(None, request.jail_info.transports)
        req.setup_jail()
        self.assertEqual([transport], request.jail_info.transports)
        req.teardown_jail()
        self.assertEqual(None, request.jail_info.transports)


class TestJailHook(TestCaseWithMemoryTransport):

    def setUp(self):
        super(TestJailHook, self).setUp()

        def clear_jail_info():
            request.jail_info.transports = None
        self.addCleanup(clear_jail_info)

    def test_jail_hook(self):
        request.jail_info.transports = None
        _pre_open_hook = request._pre_open_hook
        # Any transport is fine if jail_info.transports is None
        t = self.get_transport('foo')
        _pre_open_hook(t)
        # A transport in jail_info.transports is allowed
        request.jail_info.transports = [t]
        _pre_open_hook(t)
        # A child of a transport in jail_info is allowed
        _pre_open_hook(t.clone('child'))
        # A parent is not allowed
        self.assertRaises(errors.JailBreak, _pre_open_hook, t.clone('..'))
        # A completely unrelated transport is not allowed
        self.assertRaises(errors.JailBreak, _pre_open_hook,
                          transport.get_transport_from_url('http://host/'))

    def test_open_bzrdir_in_non_main_thread(self):
        """Opening a bzrdir in a non-main thread should work ok.

        This makes sure that the globally-installed
        breezy.bzr.smart.request._pre_open_hook, which uses a threading.local(),
        works in a newly created thread.
        """
        bzrdir = self.make_controldir('.')
        transport = bzrdir.root_transport
        thread_result = []

        def t():
            BzrDir.open_from_transport(transport)
            thread_result.append('ok')
        thread = threading.Thread(target=t)
        thread.start()
        thread.join()
        self.assertEqual(['ok'], thread_result)



# Copyright (C) 2006-2017 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Launchpad plugin commands."""

from __future__ import absolute_import

from ... import (
    branch as _mod_branch,
    controldir,
    trace,
    )
from ...commands import (
    Command,
    )
from ...errors import (
    BzrCommandError,
    NotBranchError,
    )
from ...i18n import gettext
from ...option import (
    Option,
    ListOption,
    )
from ...sixish import (
    text_type,
    )


class cmd_launchpad_open(Command):
    __doc__ = """Open a Launchpad branch page in your web browser."""

    aliases = ['lp-open']
    takes_options = [
        Option('dry-run',
               'Do not actually open the browser. Just say the URL we would '
               'use.'),
        ]
    takes_args = ['location?']

    def _possible_locations(self, location):
        """Yield possible external locations for the branch at 'location'."""
        yield location
        try:
            branch = _mod_branch.Branch.open_containing(location)[0]
        except NotBranchError:
            return
        branch_url = branch.get_public_branch()
        if branch_url is not None:
            yield branch_url
        branch_url = branch.get_push_location()
        if branch_url is not None:
            yield branch_url

    def _get_web_url(self, service, location):
        from .lp_registration import (
            InvalidURL,
            NotLaunchpadBranch)
        for branch_url in self._possible_locations(location):
            try:
                return service.get_web_url_from_branch_url(branch_url)
            except (NotLaunchpadBranch, InvalidURL):
                pass
        raise NotLaunchpadBranch(branch_url)

    def run(self, location=None, dry_run=False):
        from .lp_registration import (
            LaunchpadService)
        if location is None:
            location = u'.'
        web_url = self._get_web_url(LaunchpadService(), location)
        trace.note(gettext('Opening %s in web browser') % web_url)
        if not dry_run:
            import webbrowser   # this import should not be lazy
            # otherwise brz.exe lacks this module
            webbrowser.open(web_url)


class cmd_launchpad_login(Command):
    __doc__ = """Show or set the Launchpad user ID.

    When communicating with Launchpad, some commands need to know your
    Launchpad user ID.  This command can be used to set or show the
    user ID that Bazaar will use for such communication.

    :Examples:
      Show the Launchpad ID of the current user::

          brz launchpad-login

      Set the Launchpad ID of the current user to 'bob'::

          brz launchpad-login bob
    """
    aliases = ['lp-login']
    takes_args = ['name?']
    takes_options = [
        'verbose',
        Option('no-check',
               "Don't check that the user name is valid."),
        ]

    def run(self, name=None, no_check=False, verbose=False):
        # This is totally separate from any launchpadlib login system.
        from . import account
        check_account = not no_check

        if name is None:
            username = account.get_lp_login()
            if username:
                if check_account:
                    account.check_lp_login(username)
                    if verbose:
                        self.outf.write(gettext(
                            "Launchpad user ID exists and has SSH keys.\n"))
                self.outf.write(username + '\n')
            else:
                self.outf.write(gettext('No Launchpad user ID configured.\n'))
                return 1
        else:
            name = name.lower()
            if check_account:
                account.check_lp_login(name)
                if verbose:
                    self.outf.write(gettext(
                        "Launchpad user ID exists and has SSH keys.\n"))
            account.set_lp_login(name)
            if verbose:
                self.outf.write(gettext("Launchpad user ID set to '%s'.\n") %
                                (name,))


class cmd_launchpad_logout(Command):
    __doc__ = """Unset the Launchpad user ID.

    When communicating with Launchpad, some commands need to know your
    Launchpad user ID.  This command will log you out from Launchpad.
    This means that communication with Launchpad will happen over
    HTTPS, and will not require one of your SSH keys to be available.
    """
    aliases = ['lp-logout']
    takes_options = ['verbose']

    def run(self, verbose=False):
        from . import account
        old_username = account.get_lp_login()
        if old_username is None:
            self.outf.write(gettext('Not logged into Launchpad.\n'))
            return 1
        account.set_lp_login(None)
        if verbose:
            self.outf.write(gettext(
                "Launchpad user ID %s logged out.\n") %
                old_username)


class cmd_lp_propose_merge(Command):
    __doc__ = """Propose merging a branch on Launchpad.

    This will open your usual editor to provide the initial comment.  When it
    has created the proposal, it will open it in your default web browser.

    The branch will be proposed to merge into SUBMIT_BRANCH.  If SUBMIT_BRANCH
    is not supplied, the remembered submit branch will be used.  If no submit
    branch is remembered, the development focus will be used.

    By default, the SUBMIT_BRANCH's review team will be requested to review
    the merge proposal.  This can be overriden by specifying --review (-R).
    The parameter the launchpad account name of the desired reviewer.  This
    may optionally be followed by '=' and the review type.  For example:

      brz lp-propose-merge --review jrandom --review review-team=qa

    This will propose a merge,  request "jrandom" to perform a review of
    unspecified type, and request "review-team" to perform a "qa" review.
    """

    hidden = True
    takes_options = [Option('staging',
                            help='Propose the merge on staging.'),
                     Option('message', short_name='m', type=text_type,
                            help='Commit message.'),
                     Option('approve',
                            help=('Mark the proposal as approved immediately, '
                                  'setting the approved revision to tip.')),
                     Option('fixes', 'The bug this proposal fixes.', str),
                     ListOption('review', short_name='R', type=text_type,
                                help='Requested reviewer and optional type.')]

    takes_args = ['submit_branch?']

    aliases = ['lp-submit', 'lp-propose']

    def run(self, submit_branch=None, review=None, staging=False,
            message=None, approve=False, fixes=None):
        from . import lp_propose
        tree, branch, relpath = controldir.ControlDir.open_containing_tree_or_branch(
            '.')
        if review is None:
            reviews = None
        else:
            reviews = []
            for review in review:
                if '=' in review:
                    reviews.append(review.split('=', 2))
                else:
                    reviews.append((review, ''))
            if submit_branch is None:
                submit_branch = branch.get_submit_branch()
        if submit_branch is None:
            target = None
        else:
            target = _mod_branch.Branch.open(submit_branch)
        proposer = lp_propose.Proposer(tree, branch, target, message,
                                       reviews, staging, approve=approve,
                                       fixes=fixes)
        proposer.check_proposal()
        proposer.create_proposal()


class cmd_lp_find_proposal(Command):

    __doc__ = """Find the proposal to merge this revision.

    Finds the merge proposal(s) that discussed landing the specified revision.
    This works only if the if the merged_revno was recorded for the merge
    proposal.  The proposal(s) are opened in a web browser.

    Only the revision specified is searched for.  To find the mainline
    revision that merged it into mainline, use the "mainline" revision spec.

    So, to find the merge proposal that reviewed line 1 of README::

      brz lp-find-proposal -r mainline:annotate:README:1
    """

    takes_options = ['revision']

    def run(self, revision=None):
        from ... import ui
        from . import lp_api
        import webbrowser
        b = _mod_branch.Branch.open_containing('.')[0]
        with ui.ui_factory.nested_progress_bar() as pb, b.lock_read():
            if revision is None:
                revision_id = b.last_revision()
            else:
                revision_id = revision[0].as_revision_id(b)
            merged = self._find_proposals(revision_id, pb)
            if len(merged) == 0:
                raise BzrCommandError(gettext('No review found.'))
            trace.note(gettext('%d proposals(s) found.') % len(merged))
            for mp in merged:
                webbrowser.open(lp_api.canonical_url(mp))

    def _find_proposals(self, revision_id, pb):
        from launchpadlib import uris
        from . import lp_api
        # "devel" because branches.getMergeProposals is not part of 1.0 API.
        lp_base_url = uris.LPNET_SERVICE_ROOT
        launchpad = lp_api.connect_launchpad(lp_base_url, version='devel')
        pb.update(gettext('Finding proposals'))
        return list(launchpad.branches.getMergeProposals(
                    merged_revision=revision_id.decode('utf-8')))



# Copyright (C) 2017 Bazaar hackers
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""
Module to aid writing a Python dialect compatible with 2.7 and 3.4+.

Initially pretty much just a subset of six while things get worked out.
"""

from __future__ import absolute_import

from six import (
    binary_type,
    get_unbound_function,
    indexbytes,
    int2byte,
    PY3,
    reraise,
    string_types,
    text_type,
    unichr,
    viewitems,
    viewkeys,
    viewvalues,
    )  # noqa: F401


# The io module exists in Python 2.7 but lacks optimisation. Most uses are not
# performance critical, but want to measure before switching from cStringIO.
if PY3:
    import io as _io
    BytesIO = _io.BytesIO
    StringIO = _io.StringIO
    from builtins import range, map, zip
else:
    from cStringIO import StringIO as BytesIO  # noqa: F401
    from StringIO import StringIO  # noqa: F401
    from future_builtins import zip, map  # noqa: F401
    range = xrange  # noqa: F821


# GZ 2017-06-10: Work out if interning bits of inventory is behaviour we want
# to retain outside of StaticTuple, if so need to implement for Python 3.
if PY3:
    def bytesintern(b):
        """Dummy intern() function."""
        return b
else:
    bytesintern = intern
